#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "listes_generiques.h"

/*! \file listes_generiques.c
 * \brief L'implémentation des fonctions
 * Les définitions des fonctions sont identiques à celles de liste des entiers.
 * Mais certaines nécessitent un paramètre supplémentaire la fonction copier 
 * afficher ou detruire qui permettent de rendre générique la liste.
 * Pour l'essentiel il s'agit de reprendre les fonctions pour les listes 
 * d'entiers et de les adapter à la généricité .
 *
 * \copyright PASD
 * \version 2017
 */

typedef struct maillon_struct * maillon;

struct maillon_struct {
	void * val;
	struct maillon_struct * precedent;
	struct maillon_struct * suivant;
};

/*!
 * Création d'un nouveau maillon dont la valeur est val
 * suivant et précédent pointent sur lui même
 * \param val la valeur à stocker dans ce maillon
 * \param copier la fonction de copie correspondant au type de val
 * \return un pointeur vers le nouveau maillon
 */
static maillon maillon_creer(void * val, void(*copier)(void* val, void** pt)) {
	maillon m = malloc(sizeof(struct maillon_struct));
	(*copier)(val,&(m->val));
	m->precedent = m;
	m->suivant = m;
	return m;
}

/*!
 * Test pour savoir si le maillon est unique (bouclant sur lui-même et correspondant à une liste de taille 1).
 * \param m maillon à tester.
 */
static bool maillon_est_unique(maillon m){
	return((m->suivant == m) && (m->precedent == m));
}

/*! 
 * Destruction de toute la liste chaînée.
 * \param m (pointeur sur un) maillon de la liste chaînée à détruire.
 * \param detruire la fonction de destruction correspondant au type stocké dans m
*/
static void maillon_detruire(maillon* m, void(* detruire)(void** pt)) {
	maillon tmp = (*m)->suivant;
	(*detruire)((*m)->val);
	free(*m);
	*m = NULL;
	if(tmp) maillon_detruire(&tmp,detruire); //on detruit recursivement tous les maillons liés à m
}

/*! 
 * Affichage d'un ensemble de maillon à partir d'un début et d'une fin.
 * Le résultat est de la forme \verbatim[ 1 2 3 ]\endverbatim sans saut de ligne.
 * \param f flux où imprimer.
 * \param m_debut premier maillon à imprimer. 
 * \param m_fin dernier maillon à imprimer. 
 * \param afficher la fonction d'affichage du type correspondant
 */
static void maillon_afficher(FILE* f, maillon m_debut, maillon m_fin, void(* afficher )(FILE *f, void* val)) {
	maillon tmp = m_debut;
	while(tmp != m_fin){
		(*afficher)(f,tmp->val);
		tmp = tmp->suivant;
	}
}

/*!
 * Ajout d'un élément avant le maillon.
 * \param m maillon avant lequel on doit insérer.
 * \param _val valeur entière à insérer.
 * \param copier la fonction de copie du type correspondant
 */
static void maillon_ajouter_avant(maillon m, void* _val, void(*copier)(void* val, void** pt)) {
	maillon newM = malloc(sizeof(struct maillon_struct));
	newM->precedent = m->precedent;
	newM->suivant = m;
	(*copier)(_val,&(newM->val));
	m->precedent->suivant = newM; //on lie le nouveau maillon à son prédecesseur
	m->precedent = newM;
}

/*!
 * Ajout d'un élément apres le maillon.
 * \param m maillon apres lequel on doit insérer.
 * \param _val valeur entière à insérer.
 * \param copier la fonction de copie du type correspondant
 */
static void maillon_ajouter_apres(maillon m, void * _val, void(*copier)(void* val, void** pt)) {
	maillon newM = malloc(sizeof(struct maillon_struct));
	newM->precedent = m;
	newM->suivant = m->suivant;
	(*copier)(_val,&(newM->val));
	m->suivant->precedent = newM;
	m->suivant = newM;
}

/*! 
 * Suppression du maillon d'avant s'il existe (c.-à-d. si la liste correspondante ne contient qu'un maillon on ne fait rien).
 * \param m maillon avant lequel on doit supprimer.
 * \param detruire fonction de destruction du type correspondant.
 */
static void maillon_supprimer_avant(maillon m, void(*detruire)(void** pt)) {
	if((!m->precedent) || (m->precedent == m)) return; //si il n'y a qu'un seul maillon
	maillon nouveauPrec = m->precedent->precedent; //stockage du maillon 2 positions avant m
	nouveauPrec->suivant = m;
	(*detruire)(&(m->precedent->val));
	free(m->precedent);
	m->precedent = nouveauPrec;
}

/*! 
 * Suppression du maillon d'apres s'il existe (c.-à-d. si la liste correspondante ne contient qu'un maillon on ne fait rien).
 * \param m maillon apres lequel on doit supprimer.
 * \param detruire fonction de destruction du type correspondant.
 */
static void maillon_supprimer_apres(maillon m, void(*detruire)(void** pt)) {
	if((!m->suivant) || (m->suivant == m)) return; //si il n'y a qu'un seul maillon
	maillon nouveauSuiv = m->suivant->suivant; //stockage du maillon 2 positions apres m
	nouveauSuiv->precedent = m;
	(*detruire)(&(m->suivant->val));
	free(m->suivant);
	m->suivant = nouveauSuiv;
}


/*!
 * la structure définit 3 champs supplémentaires qui sont des pointeurs sur fonction
*/
struct liste_struct {
  unsigned int taille;
  maillon tete;
  maillon courant;
  void ( *copier ) ( void * val , void ** pt );
  void ( *afficher ) ( FILE * f , void * val );
  void ( *detruire ) ( void ** pt);

};


liste liste_creer(void(*_copier)(void* val, void** pt),
		    void(*_afficher)(FILE *f, void* val),
		    void(* _detruire )(void** pt)) {
	liste l = malloc(sizeof(struct liste_struct));
	l->taille = 0;
	l->tete = NULL;
	l->courant = NULL;
	l->copier = _copier;
	l->afficher = _afficher;
	l->detruire = _detruire;
	return l;
}


void liste_detruire(liste* l){
	free(*l);
}


void liste_insertion_debut(liste l, void* val){
	if(l->taille == 0){
	   maillon m = maillon_creer(val,l->copier);
	   l->tete = m;
	   l->taille = 1;
	   return;
	}
	maillon_ajouter_avant(l->tete,val,l->copier); //on ajoute le maillon avant la tete
	l->tete = l->tete->precedent; //on maj la nouvelle tete
	l->taille++;
}


void liste_insertion_fin(liste l, void* val){
	if(l->taille == 0){
	   maillon m = maillon_creer(val,l->copier);
	   l->tete = m;
	   l->taille = 1;
	   return;
	}
	maillon_ajouter_avant(l->tete,val,l->copier); //on ajoute le maillon avant la tete donc en fin de liste car cest une liste circulaire
	l->taille++;
}


void liste_insertion_avant(liste l, void* val){
	assert(l->courant); //courant non nul
	maillon_ajouter_avant(l->courant,val,l->copier);
	l->taille++;
}


void liste_insertion_apres(liste l, void* val){
	assert(l->courant); //courant non nul
	maillon_ajouter_apres(l->courant,val,l->copier);
	l->taille++;
}


void liste_suppression_debut(liste l){
	l->tete = l->tete->suivant;
	maillon_supprimer_avant(l->tete,l->detruire);
	l->taille--;
}


void liste_suppression_fin(liste l){
	maillon_supprimer_avant(l->tete,l->detruire);
	l->taille--;
}


void liste_suppression_avant(liste l){
	assert(l->courant); //courant non nul
	if(l->courant->precedent == l->tete) l->tete = l->courant;
	maillon_supprimer_avant(l->courant,l->detruire);
	l->taille--;
}


void liste_suppression_apres(liste l){
	assert(l->courant); //courant non nul
	if(l->courant->suivant == l->tete) l->tete = l->courant;
	maillon_supprimer_apres(l->courant,l->detruire);
	l->taille--;
}


void liste_affichage(FILE* f,liste l){
	maillon_afficher(f,l->tete,l->tete->precedent,l->afficher);
}


void liste_courant_init(liste const l) {
	l->courant = l->tete;
}


void liste_courant_suivant(liste const l) {
	l->courant = l->courant->suivant;
}


void liste_decalage(liste const l, int n) {
	if(n<0){
		for(int i = 0 ; i > n ; i--) l->courant = l->courant->precedent;
	}else if(n>0){
		for(int i = 0 ; i < n ; i++) liste_courant_suivant(l);
	}
}

void* liste_valeur_tete(liste const l) {
	return(l->tete->val);
}


void* liste_valeur_courant(liste const l) {
	return(l->courant->val);
}


bool liste_est_tete(liste const l) {
	return(l->tete == l->courant);
}

unsigned int liste_taille(liste const l) {
	return(l->taille);
}