#ifndef LISTES_GENERIQUES
#define LISTES_GENERIQUES
#include <stdbool.h>

/*! \file listes_generiques.h
 * \brief Module liste générique.
 *
 * La structure est très similaire à celle pour les entiers
 * L'objectif est de rajouter des fonctions dans cette structure
 * afin de construire une liste d'éléments de type quelconque
 *
 * \copyright PASD
 * \version 2017
 */


typedef struct liste_struct * liste;

/*! 
 * Création d'une liste vide. 
 * \param _copie la fonction de copie du type voulu.
 * \param _afficher la fonction d'affichage du type voulu.
 * \param _detruire la fonction de destruction du type voulu.
 * \return une nouvelle liste vide.
 */
liste liste_creer (void (* _copie)(void* val, void** pt), 
		    void (* _afficher)(FILE* f, void * val), 
		    void (* _detruire)(void** pt)) ;

/*!
 * Destruction de la liste.
 * \param l (pointeur vers la) liste à détruire.
 */
void liste_detruire (liste * l);

/*! 
 * Insertion en début de liste.
 * À la sortie de la fonction, tete pointe désormais sur ce nouvel élément (prendre en compte le cas de la liste vide).
 * \param l liste où insérer.
 * \param val valeur à insérer.
 */
void liste_insertion_debut (liste l, void * val);

/*! 
 * Insertion en fin de liste.
 * \param l liste où insérer.
 * \param val valeur à insérer.
 */
void liste_insertion_fin (liste l, void * val);

/*! 
 * Insertion après le maillon courant.
 * \param l liste où insérer.
 * \param val valeur à insérer.
 */
void liste_insertion_apres (liste l, void * val);

/*! 
 * Insertion avant le maillon courant.
 * Si tete = courant alors tete doit être mis à jour .
 * \param l liste où insérer.
 * \param val valeur à insérer.
 */
void liste_insertion_avant (liste l, void * val);

/*! 
 * Suppression de l'élément en tete de liste (mettre à jour tete). 
 * \param l liste d'où supprimer.
 */
void liste_suppression_debut (liste l);

/*! 
 * Suppression de l'élément en pied de liste.
 * \param l liste d'où supprimer.
 */
void liste_suppression_fin (liste l);

/*! 
 * Suppression de l'élément avant courant s'il existe (tete et pied doivent être modifiés si nécessaire).
 * \param l liste d'où supprimer.
 */
void liste_suppression_avant (liste l);

/*! 
 * Suppression de l'élément apres courant s'il existe (tete et pied doivent être modifiés si nécessaire).
 * \param l liste d'où supprimer.
 */
void liste_suppression_apres (liste l);

/*! 
 * Affichage des éléments de la liste.
 * \param f flux où imprimer.
 * \param l liste à afficher.
 */
void liste_affichage (FILE * f, liste l);

/*! 
 * Initialise courant à tete.
 * \param l liste à parcourir.
 */
void liste_courant_init (liste l);

/*!
 * Passe courant sur l'élément suivant.
 * \param l liste parcourue.
 */
void liste_courant_suivant (liste l);

/*! 
 * Décalage de courant de n positions.
 * Si n est négatif, il s'agit de décalage dans l'autre sens.
 * \param n nombre de décalage à faire.
 * \param l liste parcourue.
*/
void liste_decalage (liste l, int n);

/*!
 * Retourne la taille de la liste.
 * \param l liste sur laquelle porte la demande.
 * \return le nombre d'éléments dans la liste.
 */
unsigned int liste_taille (liste l);

/*!
 * Retourne la valeur de tete.
 * \param l liste sur laquelle porte la demande.
 * \return le nombre d'éléments dans la liste.
 */
void* liste_valeur_tete (liste l);

/*!
 * Retourne la valeur de courant.
 * \param l liste parcourue sur laquelle porte la demande.
 * \return la value de l'élément courant de la liste.
 */
void* liste_valeur_courant (liste l);

/*!
 * Retourne vrai si courant pointe sur tete. 
 * \param l liste sur laquelle porte la demande.
 * \return true si la tete est l'élément courant.
 */
bool liste_est_tete (liste l);


#endif
