#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "listes_int.h"


/*! \file listes_int.c
 * \brief La documentation pour les parties (types, variables, fonctions…) internes du module sont documentées ici.
 *
 * \copyright PASD
 * \version 2017
 */


/*!
 * Structure servant à définir un élément de type int de la liste.
 * Cette structure n'est visible que depuis listes_int.c .
 * Les fonctions attenantes sont static pour être également masquées. */
struct maillon_struct {
	int val;
	struct maillon_struct * precedent;
	struct maillon_struct * suivant;
};


/*! Un maillon est un pointeur sur (référence vers) une struct maillon_struct . */
typedef struct maillon_struct * maillon;


/*!
 * Création d'un maillon dont la valeur est _val .
 * suivant et precedent doivent pointer sur ce maillon.
 * \param _val valeur à stocker dans le maillon.
 * \return nouveau maillon stockant _val et bouclant sur lui-même.
 */
static maillon maillon_creer(int const _val) {
	maillon m = malloc(sizeof(struct maillon_struct));
	m->val = _val;
	m->precedent = m;
	m->suivant = m;
	return m;
}


/*!
 * Test pour savoir si le maillon est unique (bouclant sur lui-même et correspondant à une liste de taille 1).
 * \param m maillon à tester.
 */
static bool maillon_est_unique(maillon const m) {
	return((m->suivant == m) && (m->precedent == m));
}


/*! 
 * Destruction de toute la liste chaînée.
 * \param m (pointeur sur un) maillon de la liste chaînée à détruire.
*/
static void maillon_detruire(maillon * const m) {
	maillon tmp = (*m)->suivant;
	free(*m);
	*m = NULL;
	if(tmp) maillon_detruire(&tmp); //on detruit recursivement tous les maillons liés à m
}


/*! 
 * Affichage d'un ensemble de maillon à partir d'un début et d'une fin.
 * Le résultat est de la forme \verbatim[ 1 2 3 ]\endverbatim sans saut de ligne.
 * \param f flux où imprimer.
 * \param m_debut premier maillon à imprimer. 
 * \param m_fin dernier maillon à imprimer. 
 */
static void maillon_afficher(FILE * const f, maillon const m_debut,maillon const m_fin) {
	maillon tmp = m_debut;
	fprintf(f, "[ ");
	while(tmp != m_fin){
		fprintf(f, "%d ",tmp->val);
		tmp = tmp->suivant;
	}
	fprintf(f, "%d ]",tmp->val);
}


/*!
 * Ajout d'un élément avant le maillon.
 * \param m maillon avant lequel on doit insérer.
 * \param _val valeur entière à insérer.
 */
static void maillon_ajouter_avant(maillon const m, int const _val) {
	maillon newM = malloc(sizeof(struct maillon_struct));
	newM->precedent = m->precedent;
	newM->suivant = m;
	newM->val = _val;
	m->precedent->suivant = newM; //on lie le nouveau maillon à son prédecesseur
	m->precedent = newM;
}


/*!
 * Ajout d'un élément après le maillon.
 * \param m maillon après lequel on doit insérer.
 * \param _val valeur entière à insérer.
 */
static void maillon_ajouter_apres(maillon const m, int const _val) {
	maillon newM = malloc(sizeof(struct maillon_struct));
	newM->precedent = m;
	newM->suivant = m->suivant;
	newM->val = _val;
	m->suivant->precedent = newM;
	m->suivant = newM;
}


/*! 
 * Suppression du maillon d'avant s'il existe (c.-à-d. si la liste correspondante ne contient qu'un maillon on ne fait rien).
 * \param m maillon avant lequel on doit supprimer.
 */
static void maillon_supprimer_avant(maillon const m) {
	if((!m->precedent) || (m->precedent == m)) return; //si il n'y a qu'un seul maillon
	maillon nouveauPrec = m->precedent->precedent; //stockage du maillon 2 positions avant m
	nouveauPrec->suivant = m;
	free(m->precedent);
	m->precedent = nouveauPrec;
}


/*! suppression du maillon d'après s'il existe (c.-à-d. si la liste correspondante ne contient qu'un maillon on ne fait rien).
 * \param m maillon après lequel on doit supprimer.
*/
static void maillon_supprimer_apres(maillon const m) {
	if((!m->suivant) || (m->suivant == m)) return; //si il n'y a qu'un seul maillon
	maillon nouveauSuiv = m->suivant->suivant; //stockage du maillon 2 positions apres m
	nouveauSuiv->precedent = m;
	free(m->suivant);
	m->suivant = nouveauSuiv;
}


/*! 
 * Structure pour la liste en étendant la structure maillon. 
 * Quand la liste est vide, il n'y a pas de liste circulaire (et les pointeurs concernés valent NULL). 
*/
struct liste_struct {
	unsigned int taille;
	maillon tete;
	maillon courant;
};


liste liste_creer(void) {
	liste l = malloc(sizeof(struct liste_struct));
	l->taille = 0;
	l->tete = NULL;
	l->courant = NULL;
	return l;
}

bool liste_est_vide(liste const l) {
	return(l->taille == 0);
}

void liste_detruire(liste * const l) {
	free(*l);
}


/*!
 * Pour ajouter une valeur à une liste vide.
 * fonction facultative mais qui peut aider
 * \param l liste où ajouter.
 * \param val valeur entière à jouter.
 */
static void liste_ajouter_a_vide(liste const l, int const val) {
	maillon m = maillon_creer(val);
	l->tete = m;
	l->taille = 1;
}


/*! 
 * Pour enlever la dernière valeur. 
 * fonction facultative mais qui peut aider
 * \param l liste où enlever
 */
//ne fait rien si liste vide
/*static void liste_enlever_dernier(liste const l) {

}*/

void liste_insertion_debut(liste const l, int const val) {
	if(liste_est_vide(l)){
	 liste_ajouter_a_vide(l,val);
	 return;
	}
	maillon_ajouter_avant(l->tete,val); //on ajoute le maillon avant la tete
	l->tete = l->tete->precedent; //on maj la nouvelle tete
	l->taille++;
}


void liste_insertion_fin(liste const l, int const val) {
	if(liste_est_vide(l)){
	 liste_ajouter_a_vide(l,val);
	 return;
	}
	maillon_ajouter_avant(l->tete,val); //on ajoute le maillon avant la tete donc en fin de liste car cest une liste circulaire
	l->taille++;
}


void liste_insertion_apres(liste const l, int const val) {
	assert(l->courant); //courant non nul
	maillon_ajouter_apres(l->courant,val);
	l->taille++;
}


void liste_insertion_avant(liste const l, int const val) {
	assert(l->courant); //courant non nul
	maillon_ajouter_avant(l->courant,val);
	if(l->courant == l->tete) l->tete = l->tete->precedent; //maj de la tete
	l->taille++;
}


void liste_suppression_debut(liste const l) {
	l->tete = l->tete->suivant;
	maillon_supprimer_avant(l->tete);
	l->taille--;
}


void liste_suppression_fin(liste const l) {
	maillon_supprimer_avant(l->tete);
	l->taille--;
}


void liste_suppression_avant(liste const l) {
	assert(l->courant); //courant non nul
	if(l->courant->precedent == l->tete) l->tete = l->courant;
	maillon_supprimer_avant(l->courant);
	l->taille--;
}


void liste_suppression_apres(liste const l) {
	assert(l->courant); //courant non nul
	if(l->courant->suivant == l->tete) l->tete = l->courant;
	maillon_supprimer_apres(l->courant);
	l->taille--;
}


void liste_affichage(FILE * const f, liste const l) {
	fprintf(f, "Liste de %d éléments : ",l->taille);
	maillon_afficher(f,l->tete,l->tete->precedent);
}


void liste_courant_init(liste const l) {
	l->courant = l->tete;
}


void liste_courant_suivant(liste const l) {
	l->courant = l->courant->suivant;
}


void liste_decalage(liste const l, int n) {
	if(n<0){
		for(int i = 0 ; i > n ; i--) l->courant = l->courant->precedent;
	}else if(n>0){
		for(int i = 0 ; i < n ; i++) liste_courant_suivant(l);
	}
}


int liste_taille(liste const l) {
	return(l->taille);
}


int liste_valeur_tete(liste const l) {
	return(l->tete->val);
}


int liste_valeur_courant(liste const l) {
	return(l->courant->val);
}


bool liste_est_tete(liste const l) {
	return(l->tete == l->courant);
}


bool liste_est_fin (liste const l) {
	return(l->tete->precedent == l->courant);
}

